#! /usr/bin/guile \
--debug -e main -s
!#

;;;;
;;;; yar -- Yet another renumbering script
;;;;
;;;; Copyright 2006 Dean Ferreyra, All rights reserved
;;;;
;;;; This file is part of octw-guile-geda.
;;;;
;;;; octw-guile-geda is free software: you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation, either version 3 of the License, or
;;;; (at your option) any later version.
;;;; 
;;;; octw-guile-geda is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with octw-guile-geda.  If not, see <http://www.gnu.org/licenses/>.
;;;;
;;;; $Id: yar.scm,v 1.10 2008-09-05 06:32:34 dean Exp $
;;;; Dean Ferreyra


;;;; Besides your everyday renumbering, yar also substitutes keys in
;;;; curly braces with values.  Some predefined keys are {FILE},
;;;; {PAGE}, and {PAGES}.  The user can add new key/value pairs on the
;;;; command line.
;;;;
;;;; Yar also looks for keys in curly braces in refdes between the
;;;; uppercase letters portion and the question mark.  The first time
;;;; it encounters such a key, the refdes gets renumbered as usual,
;;;; e.g., "U{ATmega162}?" might become "U101".  Every subsequent
;;;; refdes with the same key will get the same renumbered refdes;
;;;; e.g., every subsequent "U{ATmega162}?" encountered by yar will be
;;;; substituted with "U101".  The initial uppercase letters become
;;;; part of the key, so "U{ATmega162}?" is distinct from
;;;; "C{ATmega162}?".  Note that the refdes values associated with the
;;;; keys are available to all schematics processed by the same
;;;; command.  One consequence of this is that a component may get a
;;;; refdes that looks like it belongs on another page.

(use-modules (srfi srfi-1)
             (srfi srfi-19))
(use-modules (ice-9 regex))
(use-modules (ice-9 getopt-long))
(use-modules (octw conditional) (octw anaphoric))
(use-modules (octw file-system) (octw gaf))

(define *cmd* "yar.scm")

(define (usage)
  (format #t "usage: ~A [OPTIONS]... <gschem-file> ...~%" *cmd*)
  (format #t "  -f, --force                force renumbering~%")
  (format #t (string-append
              "  -s, --step=NUMBER          "
              "page offset step number, defaults to 100~%"))
  (format #t (string-append
              "  -e, --extention=EXTENSION  "
              "extention of output files; defaults to \".renum\"~%"))
  (format #t (string-append
              "  -S, --subst=KEY:VALUE  "
              "add a key/value pair to the substitution list~%"))
  (format #t "  --debug                    enable debug mode~%")
  (format #t "  -h, --help                 this usage message~%")
  (exit 0))

(define (subst-pred value-or-list)
  (if (list? value-or-list)
      (for-each (lambda (str)
                  (add-subst str))
                value-or-list)
      (add-subst value-or-list))
  #t)

(define option-spec
  `((extension (single-char #\e) (value #t))
    (force (single-char #\f))
    (step (single-char #\s) (value #t))
    (subst (single-char #\S) (value #t) (predicate ,subst-pred))
    (debug)
    (help (single-char #\h))))

(define *debug?* #f)
(define *force?* #f)
(define *run-date* (date->string (current-date) "~1"))
(define *subst*
  `((FILE . ,(lambda (file-name page pages) file-name))
    (PAGES . ,(lambda (file-name page pages) pages))
    (PAGE . ,(lambda (file-name page pages) page))
    (DATE . ,(lambda (file-name page pages) *run-date*))))

(define (add-subst str)
  (let ((key-value-list (string-split str #\=)))
    (if (not (= (length key-value-list) 2))
        #f
        (set! *subst*
              (cons (cons (first key-value-list)
                          (second key-value-list))
                    *subst*)))))

(define (main args)
  (let ((options (getopt-long (command-line) option-spec)))
    (when (option-ref options 'help #f)
      (usage))
    (let ((files (option-ref options '() '())))
      (when (zero? (length files))
        (usage))
      (set! *debug?* (option-ref options 'debug #f))
      (set! *force?* (option-ref options 'force #f))
      (run files
           (option-ref options 'step 100)
           (option-ref options 'extension ".renum")))))

(define *name=>next-number* #f)
(define *refdes-key=>refdes* (make-hash-table 23))

(define (run files step ext)
  (let ((pages (length files)))
    (do ((files files (cdr files))
         (offset step (+ offset step))
         (page 1 (+ page 1)))
        ((null? files))
      (let* ((in-file (car files))
             (out-file (string-append in-file ext)))
        (when *debug?*
          (format #t "~A: processing in-file: ~S, out-file: ~S~%"
                  *cmd* in-file out-file))
        (call-with-input-and-output-file in-file out-file
          (lambda (in out)
            (set! *name=>next-number* (make-hash-table 23))
            (let ((subst (preprocess-subst *subst* out-file page pages)))
              (line-by-line
               in
               (lambda (line line-number k)
                 (display (process-subst
                           (process-refdes line offset)
                           subst offset) out))))))))))

(define *refdes-rx*
  (make-regexp
   (string-append "^([[:space:]]*refdes=)"
                  "([[:upper:]]+)"
                  "([{](.*)[}])?"
                  "([?]|[[:digit:]]+)"
                  "([[:space:]]*)$")))

(define (process-refdes line offset)
  (or
   (let ((match (regexp-exec *refdes-rx* line)))
     (and match
          (let ((name (match:substring match 2))
                (key (match:substring match 4))
                (number (match:substring match 5)))
            (let ((next-number (hash-ref *name=>next-number*
                                         name (+ offset 1))))
              (cond ((or (string=? number "?") *force?*)
                     (let ((refdes (gen-refdes! offset name key)))
                       (regexp-substitute #f match
                                          'pre 1 refdes 6 'post)))
                    (else
                     line))))))
   line))

(define (gen-next-number! offset name)
  (let ((next-number (hash-ref *name=>next-number*
                               name (+ offset 1))))
    (hash-set! *name=>next-number* name (+ 1 next-number))
    next-number))

(define (gen-refdes! offset name key)
  (let ((key (and key (string-append name "|" key))))
    (aif (and key (hash-ref *refdes-key=>refdes* key))
         it
         (let ((next-number (gen-next-number! offset name)))
           (let ((refdes (string-append name
                                        (number->string next-number))))
             (when key
               (hash-set! *refdes-key=>refdes* key refdes))
             refdes)))))

(define (preprocess-subst subst file-name page pages)
  (let ((page (number->string page))
        (pages (number->string pages)))
    (map (lambda (pair)
           (let ((id (car pair)))
             (cond ((symbol? id)
                    (let ((fn (cdr pair)))
                      `(,(string-append "[{]" (symbol->string id) "[}]")
                        . ,(fn file-name page pages))))
                   (else
                    `(,(string-append "[{]" id "[}]")
                      . ,(cdr pair))))))
         subst)))

(define (process-subst line subst offset)
  (let loop ((subst subst)
             (line line))
    (if (null? subst)
        line
        (loop (cdr subst)
              (regexp-substitute/global #f
                                        (caar subst)
                                        line
                                        'pre
                                        (cdar subst)
                                        'post)))))
